#include <fxsaber\MultiTester\MTTester.mqh> // https://www.mql5.com/ru/code/26132

#include <fxsaber\Expert.mqh> // https://www.mql5.com/ru/code/19003

#include "ValidateBase.mqh"

class VALIDATETASKS : public VALIDATEBASE
{
private:
  static string GetMyName( void )
  {
    return(::StringSubstr(::MQLInfoString(MQL_PROGRAM_PATH), ::StringLen(::TerminalInfoString(TERMINAL_DATA_PATH) + "\\MQL5\\Experts\\")));
  }

  static bool IsDirectoryExist( const string FolderName )
  {
    ::ResetLastError();

    return(!::FileIsExist(FolderName) && (::GetLastError() == ERR_FILE_IS_DIRECTORY));
  }

  static string GetExpertData( const ulong Chart = 0 )
  {
    string Str = NULL;

    MqlParam Parameters[];
    string Names[];

    if (EXPERT::Parameters(Chart, Parameters, Names, false))
    {
      Str += Parameters[0].string_value + "\n";

      const int Amount = ::ArraySize(Names);

      for (int i = 0; i < Amount; i++)
        Str += "; " + Names[i] + " = " + Parameters[i + 1].string_value + "\n";
    }

    return(Str);
  }

  static int GetLastPos( const string &Str, const short Char )
  {
    int Pos = ::StringLen(Str) - 1;

    while ((Pos >= 0) && (Str[Pos] != Char))
      Pos--;

    return(Pos);
  }

  static string GetName( const string &Path )
  {
    const int BeginPos = VALIDATETASKS::GetLastPos(Path, '\\') + 1;
    const int LastPos = VALIDATETASKS::GetLastPos(Path, '.');

    return((LastPos > BeginPos) ? ::StringSubstr(Path, BeginPos, LastPos - BeginPos) : NULL);
  }

  static ENUM_DAY_OF_WEEK GetDayOfWeek( const datetime time )
  {
    MqlDateTime sTime = {0};

    ::TimeToStruct(time, sTime);

    return((ENUM_DAY_OF_WEEK)sTime.day_of_week);
  }

  static bool IsHoliday( const datetime Date )
  {
    const ENUM_DAY_OF_WEEK DayOfWeek = VALIDATETASKS::GetDayOfWeek(Date);

    return((DayOfWeek == SUNDAY) || (DayOfWeek == SATURDAY));
  }

#define DAY (24 * 3600)

  static datetime GetPrevDate( const datetime ToDate, const int AmountDays, const bool WithHolidays = true )
  {
    datetime Date = WithHolidays ? ToDate - AmountDays * DAY : ToDate;

    if (!WithHolidays)
    {
      const int Day = AmountDays > 0 ? DAY : -DAY;
      const int AmountDays2 = ::MathAbs(AmountDays);

      for (int i = 0; i < AmountDays2;)
        if (!IsHoliday(Date -= Day))
          i++;
    }

    return(Date);
  }

  static string IntervalToString( const datetime FromDate, const datetime ToDate )
  {
    return(::TimeToString(FromDate, TIME_DATE) + "-" + ::TimeToString(ToDate, TIME_DATE) +
            " (" + (string)((int)(ToDate - FromDate) / DAY) + ")");
  }

  static datetime GetLastSymbolTime( const string Symb )
  {
    datetime Res = (datetime)::SymbolInfoInteger(Symb, SYMBOL_TIME);
    datetime Times[];

    if ((::CopyTime(Symb, PERIOD_M1, 0, 1, Times) == 1) && (Times[0] > Res))
      Res = Times[0];

    return(Res);
  }

public:
  static string Create( const uint InSampleDays, const uint OutSampleDays, const string FolderName = NULL, const bool WithHolidays = true )
  {
    string Path = NULL;

    string Str;

    if (MTTESTER::GetSettings2(Str))
    {
      string StrTmp = Str;

    #define MACROS(TYPE,VALUE) const TYPE VALUE = (TYPE)VALIDATETASKS::StringBetween(StrTmp, #VALUE + "=", "\r\n");
      MACROS(string, Expert)
      MACROS(string, Symbol)
      MACROS(int, Optimization)
      MACROS(int, Model)
      MACROS(datetime, FromDate)
      MACROS(datetime, ToDate)
      MACROS(int, Visual)
    #undef MACROS

      const datetime StopDate = VALIDATETASKS::GetPrevDate(ToDate, -(int)OutSampleDays, WithHolidays);

      int Amount = 0;

      const datetime SymbolDate = VALIDATETASKS::GetLastSymbolTime(Symbol);

    #define OPTIMIZATION_SYMBOLS 3
    #define MODEL_MATH 3
      bool ReCalculate = !Visual && !(ToDate % DAY) && (Model != MODEL_MATH) &&
                         (FromDate < ToDate) && (FromDate < SymbolDate) && ::SymbolInfoInteger(Symbol, SYMBOL_SELECT) &&
                         (Optimization != OPTIMIZATION_SYMBOLS) && Optimization &&

                         (Expert != VALIDATETASKS::GetMyName());

      if (!ReCalculate)
      {
        if (Visual)
          ::Alert("You cannot create tasks for the visual mode.");

        if (Optimization == OPTIMIZATION_SYMBOLS)
          ::Alert("You cannot create tasks for optimization for all symbols.");
        else if (!Optimization)
          ::Alert("You cannot create tasks without optimization.");

        if ((bool)(ToDate % DAY))
          ::Alert("You cannot create tasks without selecting an interval.");

        if (Model == MODEL_MATH)
          ::Alert("You cannot create tasks for the mode of mathematical calculations.");

        if (FromDate >= ToDate)
          ::Alert("FromDate (" + ::TimeToString(FromDate, TIME_DATE) + ") >= ToDate (" + ::TimeToString(FromDate, TIME_DATE) + ") - Error.");

        if (FromDate >= SymbolDate)
          ::Alert("FromDate (" + ::TimeToString(FromDate, TIME_DATE) + ") >= SymbolDate (" + ::TimeToString(SymbolDate, TIME_DATE) + ") - Error.");

        if (!::SymbolInfoInteger(Symbol, SYMBOL_SELECT))
          ::Alert("The symbol " + Symbol + " is not found.");

        if (Expert == VALIDATETASKS::GetMyName())
          ::Alert("You cannot create tasks for yourself: " + Expert);
      }
    #undef MODEL_MATH
    #undef OPTIMIZATION_SYMBOLS

      for (datetime EndDate = VALIDATETASKS::GetPrevDate(FromDate, -(int)OutSampleDays, WithHolidays);
           (EndDate < StopDate) && ReCalculate;
           EndDate = VALIDATETASKS::GetPrevDate(EndDate, -(int)OutSampleDays, WithHolidays))
      {
        const datetime BeginDate = VALIDATETASKS::GetPrevDate(EndDate, OutSampleDays, WithHolidays);
        const datetime BeginDateIS = VALIDATETASKS::GetPrevDate(BeginDate, InSampleDays, WithHolidays);

        VALIDATETASKS::StringReplace(Str, "FromDate=", "\r\n", ::TimeToString(BeginDateIS, TIME_DATE));
        VALIDATETASKS::StringReplace(Str, "ToDate=", "\r\n", ::TimeToString(BeginDate, TIME_DATE));

        const string Str2 = "; OutSample: " + VALIDATETASKS::IntervalToString(BeginDate, ::MathMin(EndDate, ToDate)) + "\n\n" +
                            "; Created by " + VALIDATETASKS::GetExpertData() + "\n" + Str;
        uchar Bytes[];

        if (::StringToCharArray(Str2, Bytes))
        {
          if (Path == NULL)
          {
            Path = FolderName + "\\" + VALIDATETASKS::GetName(Expert) + "\\" + Symbol + "\\" +
                   VALIDATETASKS::IntervalToString(FromDate, ToDate);

            int i = 0;

            while (!::IsStopped() && VALIDATETASKS::IsDirectoryExist(Path + "_" + ::IntegerToString(i, 2, '0') + "\\"))
              i++;

            if (i)
            {
              const int Answer = ::MessageBox("Continue calculations in this folder (YES)\n" +
                                              Path + "_" + ::IntegerToString(i - 1, 2, '0') +
                                              "\\\n\n or create (NO) new tasks?", __FILE__, MB_YESNOCANCEL);

              ReCalculate = (Answer == IDNO);

              if (Answer != IDCANCEL)
                Path += "_" + ::IntegerToString(i - (Answer == IDYES), 2, '0') + "\\";
              else
                Path = NULL;
            }
            else
              Path += "_00\\";
          }

          if (ReCalculate)
          {
            const string FileName = Path + ::IntegerToString(++Amount, 3, '0') + "_" +
                                    VALIDATETASKS::IntervalToString(BeginDateIS, BeginDate) + "_opt.ini";

            ::FileSave(FileName, Bytes);
          }
        }
      }
    }

    return(Path);
  }
#undef DAY
};